! WHIZARD 2.2.0_alpha-3 Oct 15 2013
! 
! Copyright (C) 1999-2013 by 
!     Wolfgang Kilian <kilian@physik.uni-siegen.de>
!     Thorsten Ohl <ohl@physik.uni-wuerzburg.de>
!     Juergen Reuter <juergen.reuter@desy.de>
!     
!     with contributions by 
!     Christian Speckner <cnspeckn@googlemail.com> 
!     and  Felix Braam, Sebastian Schmidt, Daniel Wiesler 
!
! WHIZARD is free software; you can redistribute it and/or modify it
! under the terms of the GNU General Public License as published by 
! the Free Software Foundation; either version 2, or (at your option)
! any later version.
!
! WHIZARD is distributed in the hope that it will be useful, but
! WITHOUT ANY WARRANTY; without even the implied warranty of
! MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the 
! GNU General Public License for more details.
!
! You should have received a copy of the GNU General Public License
! along with this program; if not, write to the Free Software
! Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
!
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
! This file has been stripped of most comments.  For documentation, refer
! to the source 'whizard.nw'

module events
  
  use kinds, only: default !NODEP!
  use iso_varying_string, string_t => varying_string !NODEP!
  use file_utils !NODEP!
  use diagnostics !NODEP!
  use unit_tests
  use os_interface
  
  use ifiles
  use lexers
  use parser

  use subevents
  use variables
  use expressions
  use models

  use state_matrices
  use particles
  use interactions
  use subevt_expr
  use processes

  implicit none
  private

  public :: event_t
  public :: pacify
  public :: events_test

  type :: event_config_t
     logical :: unweighted = .false.
     integer :: factorization_mode = FM_IGNORE_HELICITY
     logical :: keep_correlations = .false.
     type(parse_node_t), pointer :: pn_selection => null ()
     type(parse_node_t), pointer :: pn_reweight => null ()
     type(parse_node_t), pointer :: pn_analysis => null ()
   contains
     procedure :: write => event_config_write
  end type event_config_t

  type :: event_t
     type(event_config_t) :: config
     type(process_t), pointer :: process => null ()
     type(process_instance_t), pointer :: instance => null ()
     integer :: selected_i_mci = 0
     integer :: selected_i_term = 0
     integer :: selected_channel = 0
     logical :: is_complete = .false.
     logical :: particle_set_exists = .false.
     type(particle_set_t) :: particle_set
     logical :: sqme_is_known = .false.
     real(default) :: sqme = 0
     real(default) :: weight = 0
     type(event_expr_t) :: expr
     logical :: selection_evaluated = .false.
     logical :: passed = .false.
     real(default) :: reweight = 1
     logical :: analysis_flag = .false.
   contains
     procedure :: final => event_final
     procedure :: write => event_write
     procedure :: basic_init => event_init
     procedure :: connect => event_connect
     procedure :: set_selection => event_set_selection
     procedure :: set_reweight => event_set_reweight
     procedure :: set_analysis => event_set_analysis
     procedure :: setup_expressions => event_setup_expressions
     procedure :: factorize_process => event_factorize_process
     procedure :: evaluate_expressions => event_evaluate_expressions
     procedure :: reset => event_reset
     procedure :: generate => event_generate
     procedure :: accept_sqme => event_accept_sqme
     procedure :: set_sqme => event_set_sqme
     procedure :: select => event_select
     procedure :: set_particle_set => event_set_particle_set
     procedure :: accept_particle_set => event_accept_particle_set
     procedure :: recalculate => event_recalculate
     procedure :: get_process_ptr => event_get_process_ptr
     procedure :: get_sqme => event_get_sqme
     procedure :: get_weight => event_get_weight
     procedure :: get_sqme_process => event_get_sqme_process
     procedure :: has_particle_set => event_has_particle_set
     procedure :: get_particle_set_ptr => event_get_particle_set_ptr
     procedure :: get_i_mci => event_get_i_mci
     procedure :: get_i_term => event_get_i_term
     procedure :: get_channel => event_get_channel
     procedure :: get_fac_scale => event_get_fac_scale
     procedure :: get_alpha_s => event_get_alpha_s
  end type event_t


  interface pacify
     module procedure pacify_event
  end interface pacify

contains
  
  subroutine event_config_write (object, unit, show_expressions)
    class(event_config_t), intent(in) :: object
    integer, intent(in), optional :: unit
    logical, intent(in), optional :: show_expressions
    integer :: u
    u = output_unit (unit)
    write (u, "(3x,A,L1)")  "Unweighted         = ", object%unweighted
    write (u, "(3x,A)", advance="no")  "Helicity handling  = "
    select case (object%factorization_mode)
    case (FM_IGNORE_HELICITY)
       write (u, "(A)")  "drop"
    case (FM_SELECT_HELICITY)
       write (u, "(A)")  "select"
    case (FM_FACTOR_HELICITY)
       write (u, "(A)")  "factorize"
    end select
    write (u, "(3x,A,L1)")  "Keep correlations  = ", object%keep_correlations
    if (present (show_expressions)) then
       if (show_expressions) then
          if (associated (object%pn_selection)) then
             call write_separator (u)
             write (u, "(3x,A)") "Event selection expression:"
             call object%pn_selection%write (u)
          end if
          if (associated (object%pn_reweight)) then
             call write_separator (u)
             write (u, "(3x,A)") "Event reweighting expression:"
             call object%pn_reweight%write (u)
          end if
          if (associated (object%pn_analysis)) then
             call write_separator (u)
             write (u, "(3x,A)") "Analysis expression:"
             call object%pn_analysis%write (u)
          end if
       end if
    end if
  end subroutine event_config_write
  
  subroutine event_final (object)
    class(event_t), intent(inout) :: object
    call particle_set_final (object%particle_set)
  end subroutine event_final
    
  subroutine event_write (object, unit, verbose)
    class(event_t), intent(in) :: object
    integer, intent(in), optional :: unit
    logical, intent(in), optional :: verbose
    logical :: verb
    integer :: u
    u = output_unit (unit)
    verb = .false.;  if (present (verbose))  verb = verbose
    call write_separator_double (u)
    if (object%is_complete) then
       write (u, "(1x,A)")  "Event"
    else
       write (u, "(1x,A)")  "Event [incomplete]"
    end if
    call write_separator (u)
    call object%config%write (u)
    if (object%sqme_is_known) then
       call write_separator (u)
       write (u, "(3x,A,ES19.12)")  "Squared matrix el. = ", object%sqme
       write (u, "(3x,A,ES19.12)")  "Event weight       = ", object%weight
    end if
    if (object%selected_i_mci /= 0) then
       call write_separator (u)
       write (u, "(3x,A,I0)")  "Selected MCI group = ", object%selected_i_mci
       write (u, "(3x,A,I0)")  "Selected term      = ", object%selected_i_term
       write (u, "(3x,A,I0)")  "Selected channel   = ", object%selected_channel
    end if
    if (object%selection_evaluated) then
       call write_separator (u)
       write (u, "(3x,A,L1)")  "Passed selection   = ", object%passed
       if (object%passed) then
          write (u, "(3x,A,ES19.12)") &
               "Reweighting factor = ", object%reweight
          write (u, "(3x,A,L1)") &
               "Analysis flag      = ", object%analysis_flag
       end if
    end if
    if (associated (object%instance)) then
       if (verb) then
          call object%instance%write (u)
       else
          call object%instance%write_header (u)
       end if
       if (object%particle_set_exists) then
          call particle_set_write (object%particle_set, u)
          call write_separator_double (u)
       end if
       if (object%expr%subevt_filled) then
          call object%expr%write (u)
          call write_separator_double (u)
       end if
    else
       call write_separator_double (u)
       write (u, "(1x,A)")  "Process instance: [undefined]"
       call write_separator_double (u)
    end if
  end subroutine event_write

  subroutine event_init (event, var_list)
    class(event_t), intent(out) :: event
    type(var_list_t), intent(in) :: var_list
    event%config%unweighted = var_list_get_lval (var_list, &
         var_str ("?unweighted"))
  end subroutine event_init
    
  subroutine event_connect (event, process_instance)
    class(event_t), intent(inout), target :: event
    type(process_instance_t), intent(in), target :: process_instance
    event%process => process_instance%process
    event%instance => process_instance
    call event%expr%setup_vars (event%process%get_sqrts ())
    call event%expr%link_var_list (event%process%get_var_list_ptr ())
  end subroutine event_connect

  subroutine event_set_selection (event, pn_selection)
    class(event_t), intent(inout) :: event
    type(parse_node_t), intent(in), pointer :: pn_selection
    event%config%pn_selection => pn_selection
  end subroutine event_set_selection

  subroutine event_set_reweight (event, pn_reweight)
    class(event_t), intent(inout) :: event
    type(parse_node_t), intent(in), pointer :: pn_reweight
    event%config%pn_reweight => pn_reweight
  end subroutine event_set_reweight

  subroutine event_set_analysis (event, pn_analysis)
    class(event_t), intent(inout) :: event
    type(parse_node_t), intent(in), pointer :: pn_analysis
    event%config%pn_analysis => pn_analysis
  end subroutine event_set_analysis
  
  subroutine event_setup_expressions (event)
    class(event_t), intent(inout), target :: event
    call event%expr%setup_selection (event%config%pn_selection)
    call event%expr%setup_analysis (event%config%pn_analysis)
    call event%expr%setup_reweight (event%config%pn_reweight)
  end subroutine event_setup_expressions
  
  subroutine event_factorize_process (event, r)
    class(event_t), intent(inout) :: event
    real(default), dimension(2), intent(in) :: r
    integer :: i_term
    type(interaction_t), pointer :: int_matrix, int_flows
    integer, dimension(:), allocatable :: beam_index
    integer, dimension(:), allocatable :: in_index
    integer :: n_in
    event%particle_set_exists = .false.
    if (event%instance%is_complete_event ()) then
       call event%instance%select_i_term (i_term)
       event%selected_i_term = i_term
       int_matrix => event%instance%get_matrix_int_ptr (i_term)
       int_flows  => event%instance%get_flows_int_ptr (i_term)
       n_in = event%process%get_n_in ()
       call particle_set_init (event%particle_set, event%particle_set_exists, &
            int_matrix, int_flows, event%config%factorization_mode, r, &
            event%config%keep_correlations, keep_virtual=.true.)
       allocate (beam_index (n_in))
       call event%instance%get_beam_index (i_term, beam_index)
       call particle_set_reset_status (event%particle_set, &
            beam_index, PRT_BEAM)
       allocate (in_index (n_in))
       call event%instance%get_in_index (i_term, in_index)
       call particle_set_reset_status (event%particle_set, &
            in_index, PRT_INCOMING)
       event%particle_set_exists = .true.
    end if
    event%is_complete = event%sqme_is_known .and. event%particle_set_exists
  end subroutine event_factorize_process
    
  subroutine event_evaluate_expressions (event)
    class(event_t), intent(inout) :: event
    if (event%particle_set_exists) then
       call event%expr%fill_subevt (event%particle_set)
       call event%expr%evaluate &
            (event%passed, event%reweight, event%analysis_flag)
       event%selection_evaluated = .true.
    else
       ! error?
    end if
  end subroutine event_evaluate_expressions
  
  subroutine event_reset (event)
    class(event_t), intent(inout) :: event
    event%selected_i_mci = 0
    event%selected_i_term = 0
    event%selected_channel = 0
    event%is_complete = .false.
    event%particle_set_exists = .false.
    event%sqme_is_known = .false.
    call event%expr%reset ()
    event%selection_evaluated = .false.
    event%passed = .false.
    event%analysis_flag = .false.
    if (associated (event%instance)) then
       call event%instance%reset (reset_mci = .true.)
    end if
  end subroutine event_reset
  
  subroutine event_generate (event, i_mci, r)
    class(event_t), intent(inout) :: event
    integer, intent(in) :: i_mci
    real(default), dimension(2), intent(in) :: r
    call event%reset ()
    event%selected_i_mci = i_mci
    if (event%config%unweighted) then
       call event%process%generate_unweighted_event (event%instance, i_mci)
       call event%instance%evaluate_event_data ()
       call event%instance%normalize_weight ()
    else
       call event%process%generate_weighted_event (event%instance, i_mci)
       call event%instance%evaluate_event_data ()
    end if
    event%selected_channel = event%instance%get_channel ()
    call event%accept_sqme ()
    call event%factorize_process (r)
    call event%evaluate_expressions ()
  end subroutine event_generate
  
  subroutine event_accept_sqme (event)
    class(event_t), intent(inout) :: event
    if (associated (event%instance)) then
       if (event%instance%is_complete_event ()) then
          call event%set_sqme (event%instance%get_sqme (), &
               event%instance%get_weight ())
       end if
    end if
  end subroutine event_accept_sqme
  
  subroutine event_set_sqme (event, sqme, weight)
    class(event_t), intent(inout) :: event
    real(default), intent(in) :: sqme, weight
    event%sqme = sqme
    event%weight = weight
    event%sqme_is_known = .true.
    if (associated (event%instance)) then
       event%is_complete = event%instance%is_complete_event () &
            .and. event%sqme_is_known .and. event%particle_set_exists
    end if
  end subroutine event_set_sqme
  
  subroutine event_select (event, i_mci, i_term, channel)
    class(event_t), intent(inout) :: event
    integer, intent(in) :: i_mci, i_term, channel
    if (associated (event%instance)) then
       event%selected_i_mci = i_mci
       event%selected_i_term = i_term
       event%selected_channel = channel
    else
       call msg_bug ("Event: select term: process instance undefined")
    end if
  end subroutine event_select

  subroutine event_set_particle_set (event, pset)
    class(event_t), intent(inout) :: event
    type(particle_set_t), intent(in) :: pset
    event%particle_set = pset
    call event%accept_particle_set ()
  end subroutine event_set_particle_set

  subroutine event_accept_particle_set (event)
    class(event_t), intent(inout) :: event
    event%particle_set_exists = .true.
  end subroutine event_accept_particle_set
  
  subroutine event_recalculate (event, update_sqme)
    class(event_t), intent(inout) :: event
    logical, intent(in) :: update_sqme
    integer :: i_mci, i_term, channel
    if (event%particle_set_exists) then
       i_mci = event%selected_i_mci
       i_term = event%selected_i_term
       channel = event%selected_channel
       if (i_mci == 0 .or. i_term == 0 .or. channel == 0) then
          call msg_bug ("Event: recalculate: undefined selection parameters")
       end if
       call event%instance%choose_mci (i_mci)
       call event%instance%set_trace (event%particle_set, i_term)
       call event%instance%recover (channel, i_term, update_sqme) 
       call event%process%recover_event (event%instance, i_term)
       if (event%sqme_is_known) then
          call event%instance%evaluate_event_data (weight = event%weight)
       else
          call event%instance%evaluate_event_data ()
          if (event%config%unweighted) then
             call event%instance%normalize_weight ()
          end if
       end if
       event%is_complete = event%instance%is_complete_event () &
            .and. event%sqme_is_known
    else
       call msg_bug ("Event: can't recalculate, particle set is undefined")
    end if
  end subroutine event_recalculate
  
  function event_get_process_ptr (event) result (ptr)
    class(event_t), intent(in) :: event
    type(process_t), pointer :: ptr
    ptr => event%process
  end function event_get_process_ptr

  function event_get_sqme (event) result (sqme)
    class(event_t), intent(in) :: event
    real(default) :: sqme
    if (event%sqme_is_known) then
       sqme = event%sqme
    else
       call msg_bug ("Event: attempt to extract undefined sqme")
    end if
  end function event_get_sqme
    
  function event_get_weight (event) result (weight)
    class(event_t), intent(in) :: event
    real(default) :: weight
    if (event%sqme_is_known) then
       weight = event%weight
    else
       call msg_bug ("Event: attempt to extract undefined weight")
    end if
  end function event_get_weight
    
  function event_get_sqme_process (event) result (sqme)
    class(event_t), intent(in) :: event
    real(default) :: sqme
    if (associated (event%instance)) then
       sqme = event%instance%get_sqme ()
    else
       call msg_bug ("Event: attempt to get sqme &
            &from undefined process instance")
    end if
  end function event_get_sqme_process
  
  function event_has_particle_set (event) result (flag)
    class(event_t), intent(in) :: event
    logical :: flag
    flag = event%particle_set_exists
  end function event_has_particle_set
  
  function event_get_particle_set_ptr (event) result (ptr)
    class(event_t), intent(in), target :: event
    type(particle_set_t), pointer :: ptr
    ptr => event%particle_set
  end function event_get_particle_set_ptr
  
  function event_get_i_mci (event) result (i_mci)
    class(event_t), intent(in) :: event
    integer :: i_mci
    i_mci = event%selected_i_mci
  end function event_get_i_mci
  
  function event_get_i_term (event) result (i_term)
    class(event_t), intent(in) :: event
    integer :: i_term
    i_term = event%selected_i_term
  end function event_get_i_term
  
  function event_get_channel (event) result (channel)
    class(event_t), intent(in) :: event
    integer :: channel
    channel = event%selected_channel
  end function event_get_channel
  
  function event_get_fac_scale (event) result (fac_scale)
    class(event_t), intent(in) :: event
    real(default) :: fac_scale
    fac_scale = event%instance%get_fac_scale (event%selected_i_term)
  end function event_get_fac_scale
    
  function event_get_alpha_s (event) result (alpha_s)
    class(event_t), intent(in) :: event
    real(default) :: alpha_s
    alpha_s = event%instance%get_alpha_s (event%selected_i_term)
  end function event_get_alpha_s
    
  subroutine write_separator (u)
    integer, intent(in) :: u
    write (u, "(A)")  repeat ("-", 72)
  end subroutine write_separator
  
  subroutine write_separator_double (u)
    integer, intent(in) :: u
    write (u, "(A)")  repeat ("=", 72)
  end subroutine write_separator_double
  
  subroutine pacify_event (event)
    class(event_t), intent(inout) :: event
    if (event%particle_set_exists)  call pacify (event%particle_set)
    if (event%expr%subevt_filled)  call pacify (event%expr)
  end subroutine pacify_event
  

  subroutine events_test (u, results)
    integer, intent(in) :: u
    type(test_results_t), intent(inout) :: results
    call test (events_1, "events_1", &
         "empty event record", &
         u, results)
    call test (events_2, "events_2", &
         "generate event", &
         u, results)
    call test (events_3, "events_3", &
         "expression evaluation", &
         u, results)
    call test (events_4, "events_4", &
         "recover event", &
         u, results)
    call test (events_5, "events_5", &
         "partially recover event", &
         u, results)
  end subroutine events_test
  
  subroutine events_1 (u)
    integer, intent(in) :: u
    type(event_t), target :: event

    write (u, "(A)")  "* Test output: events_1"
    write (u, "(A)")  "*   Purpose: display an empty event object"
    write (u, "(A)")

    call event%write (u)
    
    write (u, "(A)")
    write (u, "(A)")  "* Test output end: events_1"
    
  end subroutine events_1
  
  subroutine events_2 (u)
    integer, intent(in) :: u
    type(event_t), allocatable, target :: event
    type(process_t), allocatable, target :: process
    type(process_instance_t), allocatable, target :: process_instance
    type(model_list_t) :: model_list

    write (u, "(A)")  "* Test output: events_2"
    write (u, "(A)")  "*   Purpose: generate and display an event"
    write (u, "(A)")

    call syntax_model_file_init ()

    write (u, "(A)")  "* Generate test process event"

    allocate (process)
    allocate (process_instance)
    call prepare_test_process (process, process_instance, model_list)
    call process_instance%setup_event_data ()

    write (u, "(A)")
    write (u, "(A)")  "* Initialize event object"

    allocate (event)
    call event%connect (process_instance)
    
    write (u, "(A)")
    write (u, "(A)")  "* Generate test process event"

    call process%generate_weighted_event (process_instance, 1)

    write (u, "(A)")
    write (u, "(A)")  "* Fill event object"
    write (u, "(A)")

    call event%generate (1, [0.4_default, 0.4_default])
    call event%write (u)

    write (u, "(A)")
    write (u, "(A)")  "* Cleanup"

    call event%final ()
    deallocate (event)

    call cleanup_test_process (process, process_instance)
    deallocate (process_instance)
    deallocate (process)
    
    call model_list%final ()
    call syntax_model_file_final ()
    
    write (u, "(A)")
    write (u, "(A)")  "* Test output end: events_2"
    
  end subroutine events_2
  
  subroutine events_3 (u)
    integer, intent(in) :: u
    type(string_t) :: expr_text
    type(ifile_t) :: ifile
    type(stream_t) :: stream
    type(parse_tree_t) :: pt_selection, pt_reweight, pt_analysis
    type(event_t), allocatable, target :: event
    type(process_t), allocatable, target :: process
    type(process_instance_t), allocatable, target :: process_instance
    type(model_list_t) :: model_list

    write (u, "(A)")  "* Test output: events_3"
    write (u, "(A)")  "*   Purpose: generate an event and evaluate expressions"
    write (u, "(A)")

    call syntax_model_file_init ()
    call syntax_pexpr_init ()

    write (u, "(A)")  "* Expression texts"
    write (u, "(A)")

    expr_text = "all Pt > 100 [s]"
    write (u, "(A,A)")  "selection = ", char (expr_text)
    call ifile_clear (ifile)
    call ifile_append (ifile, expr_text)
    call stream_init (stream, ifile)
    call parse_tree_init_lexpr (pt_selection, stream, .true.)
    call stream_final (stream)

    expr_text = "1 + sqrts_hat / sqrts"
    write (u, "(A,A)")  "reweight = ", char (expr_text)
    call ifile_clear (ifile)
    call ifile_append (ifile, expr_text)
    call stream_init (stream, ifile)
    call parse_tree_init_expr (pt_reweight, stream, .true.)
    call stream_final (stream)
    
    expr_text = "true"
    write (u, "(A,A)")  "analysis = ", char (expr_text)
    call ifile_clear (ifile)
    call ifile_append (ifile, expr_text)
    call stream_init (stream, ifile)
    call parse_tree_init_lexpr (pt_analysis, stream, .true.)
    call stream_final (stream)

    call ifile_final (ifile)
    
    write (u, "(A)")
    write (u, "(A)")  "* Initialize test process event"

    allocate (process)
    allocate (process_instance)
    call prepare_test_process (process, process_instance, model_list)
    call process%set_var_list &
         (model_get_var_list_ptr (process%get_model_ptr ()))
    call process_instance%setup_event_data ()

    write (u, "(A)")
    write (u, "(A)")  "* Initialize event object and set expressions"

    allocate (event)

    call event%set_selection (parse_tree_get_root_ptr (pt_selection))
    call event%set_reweight (parse_tree_get_root_ptr (pt_reweight))
    call event%set_analysis (parse_tree_get_root_ptr (pt_analysis))
    
    call event%connect (process_instance)
    call event%setup_expressions ()

    write (u, "(A)")
    write (u, "(A)")  "* Generate test process event"

    call process%generate_weighted_event (process_instance, 1)

    write (u, "(A)")
    write (u, "(A)")  "* Fill event object and evaluate expressions"
    write (u, "(A)")

    call event%generate (1, [0.4_default, 0.4_default])
    call event%write (u)

    write (u, "(A)")
    write (u, "(A)")  "* Cleanup"

    call event%final ()
    deallocate (event)

    call cleanup_test_process (process, process_instance)
    deallocate (process_instance)
    deallocate (process)
    
    call model_list%final ()
    call syntax_model_file_final ()
    
    write (u, "(A)")
    write (u, "(A)")  "* Test output end: events_3"
    
  end subroutine events_3
  
  subroutine events_4 (u)
    integer, intent(in) :: u
    type(event_t), allocatable, target :: event
    type(process_t), allocatable, target :: process
    type(process_instance_t), allocatable, target :: process_instance
    type(particle_set_t) :: particle_set
    type(model_list_t) :: model_list

    write (u, "(A)")  "* Test output: events_4"
    write (u, "(A)")  "*   Purpose: generate and recover an event"
    write (u, "(A)")

    call syntax_model_file_init ()

    write (u, "(A)")  "* Generate test process event and save particle set"
    write (u, "(A)")

    allocate (process)
    allocate (process_instance)
    call prepare_test_process (process, process_instance, model_list)
    call process_instance%setup_event_data ()

    allocate (event)
    call event%connect (process_instance)
    
    call event%generate (1, [0.4_default, 0.4_default])
    call event%write (u)
    
    particle_set = event%particle_set

    call event%final ()
    deallocate (event)

    call cleanup_test_process (process, process_instance)
    deallocate (process_instance)
    deallocate (process)

    write (u, "(A)")
    write (u, "(A)")  "* Recover event from particle set"
    write (u, "(A)")
    
    allocate (process)
    allocate (process_instance)
    call prepare_test_process (process, process_instance, model_list)
    call process_instance%setup_event_data ()

    allocate (event)
    call event%connect (process_instance)
    
    call event%select (1, 1, 1)
    call event%set_particle_set (particle_set)
    call event%recalculate (update_sqme = .true.)
    call event%write (u)

    write (u, "(A)")
    write (u, "(A)")  "* Transfer sqme and evaluate expressions"
    write (u, "(A)")
    
    call event%accept_sqme ()
    call event%evaluate_expressions ()
    call event%write (u)
    
    write (u, "(A)")
    write (u, "(A)")  "* Reset contents"
    write (u, "(A)")

    call event%reset ()
    call event%write (u)

    write (u, "(A)")
    write (u, "(A)")  "* Cleanup"

    call particle_set_final (particle_set)

    call event%final ()
    deallocate (event)

    call cleanup_test_process (process, process_instance)
    deallocate (process_instance)
    deallocate (process)
    
    call model_list%final ()
    call syntax_model_file_final ()

    write (u, "(A)")
    write (u, "(A)")  "* Test output end: events_4"
    
  end subroutine events_4
  
  subroutine events_5 (u)
    integer, intent(in) :: u
    type(event_t), allocatable, target :: event
    type(process_t), allocatable, target :: process
    type(process_instance_t), allocatable, target :: process_instance
    type(particle_set_t) :: particle_set
    real(default) :: sqme, weight
    type(model_list_t) :: model_list

    write (u, "(A)")  "* Test output: events_5"
    write (u, "(A)")  "*   Purpose: generate and recover an event"
    write (u, "(A)")

    call syntax_model_file_init ()

    write (u, "(A)")  "* Generate test process event and save particle set"
    write (u, "(A)")

    allocate (process)
    allocate (process_instance)
    call prepare_test_process (process, process_instance, model_list)
    call process_instance%setup_event_data ()

    allocate (event)
    call event%connect (process_instance)
    
    call event%generate (1, [0.4_default, 0.4_default])
    call event%write (u)
    
    particle_set = event%particle_set
    sqme = event%get_sqme ()
    weight = event%get_weight ()

    call event%final ()
    deallocate (event)

    call cleanup_test_process (process, process_instance)
    deallocate (process_instance)
    deallocate (process)

    write (u, "(A)")
    write (u, "(A)")  "* Recover event from particle set"
    write (u, "(A)")
    
    allocate (process)
    allocate (process_instance)
    call prepare_test_process (process, process_instance, model_list)
    call process_instance%setup_event_data ()

    allocate (event)
    call event%connect (process_instance)
    
    call event%select (1, 1, 1)
    call event%set_particle_set (particle_set)
    call event%recalculate (update_sqme = .false.)
    call event%write (u)

    write (u, "(A)")
    write (u, "(A)")  "* Manually set sqme and evaluate expressions"
    write (u, "(A)")
    
    call event%set_sqme (sqme, weight)
    call event%evaluate_expressions ()
    call event%write (u)
    
    write (u, "(A)")
    write (u, "(A)")  "* Cleanup"

    call particle_set_final (particle_set)

    call event%final ()
    deallocate (event)

    call cleanup_test_process (process, process_instance)
    deallocate (process_instance)
    deallocate (process)
    
    call model_list%final ()
    call syntax_model_file_final ()

    write (u, "(A)")
    write (u, "(A)")  "* Test output end: events_5"
    
  end subroutine events_5
  

end module events
